import json
import os

import anthropic
from dotenv import load_dotenv

from justai.models.model import Model, OverloadedException
from justai.tools.display import ERROR_COLOR, color_print


# Models:
# Claude 3 Opus	    claude-3-opus-20240229
# Claude 3 Sonnet	claude-3-sonnet-20240229
# Claude 3 Haiku	Coming soon
#
# Supported parameters:
# max_tokens: int (default 800)
# temperature: float (default 0.8)
#
# Asynchronous use:
# Call Agent.chat_async(messages: list[dict]) -> str
# In contrast to Agent.chat, Agent.chat_async cannot return json and does not return input and output token counts

class AnthropicModel(Model):

    def __init__(self, model_name: str, params: dict):
        """ Model implemention should create attributes for all supported parameters """
        system_message = f"You are {model_name}, a large language model trained by Anthropic."
        super().__init__(model_name, params, system_message)

        # Authentication
        if "ANTHROPIC_API_KEY" in params:
            api_key = params["ANTHROPIC_API_KEY"]
        else:
            if not os.getenv("ANTHROPIC_API_KEY"):
                load_dotenv()  # Load the .env file into the environment
            api_key = os.getenv("ANTHROPIC_API_KEY")
        if not api_key:
            color_print("No Anthropic API key found. Create one at https://console.anthropic.com/settings/keys and " +
                        "set it in the .env file like ANTHROPIC_API_KEY=here_comes_your_key.", color=ERROR_COLOR)

        # Client
        if params.get('async'):
            self.client = anthropic.AsyncAnthropic(api_key=api_key)
        else:
            self.client = anthropic.Anthropic(api_key=api_key)

        # Model specific parameters
        self.model_params['max_tokens'] = params.get('max_tokens', 800)
        self.model_params['temperature'] = params.get('temperature', 0.8)

    def chat(self, messages: list[dict], return_json: bool, max_retries=None, log=None) \
            -> tuple[[str | object], int, int]:

        messages = anthropic_messages(messages, return_json)

        try:
            message = self.client.messages.create(
                model=self.model_name,
                max_tokens=self.model_params['max_tokens'],
                temperature=self.model_params['temperature'],
                system=self.system_message,
                messages=messages
            )
        except anthropic.InternalServerError as e:
            raise OverloadedException(e)

        response = message.content[0].text
        if return_json:
            response = json.loads('{' + response)
        input_tokens = message.usage.input_tokens
        output_tokens = message.usage.output_tokens
        return response, input_tokens, output_tokens

    def chat_async(self, messages: list[dict]) -> str:

        messages = anthropic_messages(messages, return_json=False)

        try:
            stream = self.client.messages.create(
                model=self.model_name,
                max_tokens=self.model_params['max_tokens'],
                temperature=self.model_params['temperature'],
                system=self.system_message,
                messages=messages,
                stream=True
            )
        except anthropic.InternalServerError as e:
            raise OverloadedException(e)

        for event in stream:
            if hasattr(event, 'delta') and hasattr(event.delta, 'text'):
                yield event.delta.text

    def token_count(self, text: str) -> int:
        return -1  # Not implemented


def anthropic_messages(messages: list[dict], return_json: bool) -> list[dict]:
    messages = [{"role": m['role'],
                 "content": [{"type": "text",
                              "text": m['content']}]
                 } for m in messages if m['role'] != 'system']
    if return_json:
        messages += [{"role": 'assistant',
                      "content": [{"type": "text",
                                   "text": "sure here's your json: {"}]
                      }]
    return messages



import json
import os
import sqlite3
from pathlib import Path

from justai.tools.display import color_print

log_dir = ''
log_file = 'justai_log.db'
log_retention_hours = 72

COLORS = {
    "error": "#FF0000",
    "warning": "#EECC00",
    "response": "#0000FF",
    "prompt": "#008000",
    "info": "#888888",
    "default": "#666666"
}


def set_log_dir(_dir):
    global log_dir
    log_dir = _dir


class Log:
    _instance = None

    def __new__(cls, *args, **kwargs):  # Make this class a singleton
        if cls._instance is None:
            cls._instance = super(Log, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        global log_dir, log_file

        dir_ = os.getenv('LOG_DIR', log_dir) or Path(__file__).resolve().parent
        self.log_path = os.path.join(dir_, log_file)
        self.conn = sqlite3.connect(self.log_path, check_same_thread=False)
        self.cursor = self.conn.cursor()
        self.cursor.execute('''CREATE TABLE IF NOT EXISTS log (
                                    timestamp DATETIME DEFAULT CURRENT_TIMESTAMP,
                                    title TEXT,
                                    value TEXT,
                                    logtype VARCHAR(10))''')
        self.conn.commit()
        self.cleanup_logs()

    def cleanup_logs(self):
        global log_retention_hours

        sql = "DELETE FROM log WHERE timestamp < datetime('now', ? || ' hours')"
        parameters = (f'-{log_retention_hours}',)
        try:
            # Create a new cursor for the delete operation
            with self.conn:
                cursor = self.conn.cursor()
                cursor.execute(sql, parameters)
                self.conn.commit()
        except sqlite3.OperationalError as e:
            print(f"OperationalError: {e}")
        except sqlite3.ProgrammingError as e:
            print(f"ProgrammingError: {e}")
        except Exception as e:
            print(f"An error occurred: {e}")

    def write(self, title, text, logtype='default'):
        if not isinstance(text, str):
            try:
                text = json.dumps(text, indent=4)
            except Exception:
                text = str(text)
        self.conn.execute('INSERT INTO log (title, value, logtype) VALUES (?, ?, ?)', (title, text, logtype))
        try:
            self.conn.commit()
        except:
            pass # Ignore errors when the log is closed
        print('\n' + title)
        color_print(text, COLORS[logtype])

    def error(self, title, text):
        self.write(title, text, 'error')

    def info(self, title, text):
        self.write(title, text, 'info')

    def warning(self, title, text):
        self.write(title, text, 'warning')

    def prompt(self, title, text):
        self.write(title, text, 'prompt')

    def response(self, title, text):
        self.write(title, text, 'response')

    def clear(self):
        self.cursor.execute('DELETE FROM log')
        self.conn.commit()

    def is_empty(self):
        self.cursor.execute('SELECT COUNT(*) FROM log')
        return self.cursor.fetchone()[0] == 0

    def close(self):
        self.conn.close()

    def as_html(self):
        self.cursor.execute('SELECT * FROM log')
        result = ''
        for res in reversed(self.cursor.fetchall()):
            timestamp, title, value, logtype = res
            value = value.replace('\n', '<br/>')
            color = COLORS.get(logtype, 'gray')
            result += f'<p><b>{timestamp} - {title}</b><br/><span style="color:{color}">{value}</span></p>\n'
        return result


if __name__ == "__main__":
    log = Log()
    log.prompt('prompt', 'System message', )
    print(log.as_html())
    log.response('response', 'This is the response')
    log.error('Oops', 'There was an error')
    log.info('Info', 'This is some information')
    print(log.as_html())


LANGUAGES_NL = {
    "Albanees": "sq-AL",
    "Amhaars": "am-ET",
    "Arabisch": "ar-SA",
    "Armeens": "hy-AM",
    "Azerbeidzjaans": "az-AZ",
    "Baskisch": "eu-ES",
    "Bengaals": "bn-BD",
    "Bhutanees": "dz-BT",
    "Birmaans": "my-MM",
    "Bosnisch": "bs-BA",
    "Bretons": "br-FR",
    "Bulgaars": "bg-BG",
    "Catalaans": "ca-ES",
    "Chinees": "zh-CN",  # Vereenvoudigd Chinees, voor Traditioneel Chinees zou het "zh-TW" zijn
    "Corsicaans": "co-FR",
    "Deens": "da-DK",
    "Duits": "de-DE",
    "Dzongkha": "dz-BT",
    "Engels": "en-GB",
    "Engels (Brits)": "en-GB",
    "Engels (VS)": "en-US",
    "Amerikaans": "en-US",
    "Estlands": "et-EE",
    "Farsi": "fa-IR",
    "Perzisch": "fa-IR",
    "Fijisch": "fj-FJ",
    "Fins": "fi-FI",
    "Frans": "fr-FR",
    "Gaelisch (Schots)": "gd-GB",
    "Galicisch": "gl-ES",
    "Georgisch": "ka-GE",
    "Grieks": "el-GR",
    "Hebreeuws": "he-IL",
    "Hindi": "hi-IN",
    "Hongaars": "hu-HU",
    "IJslands": "is-IS",
    "Indonesisch": "id-ID",
    "Italiaans": "it-IT",
    "Japans": "ja-JP",
    "Javaans": "jv-ID",
    "Kazachs": "kk-KZ",
    "Khmer": "km-KH",
    "Koreaans": "ko-KR",
    "Kroatisch": "hr-HR",
    "Lao": "lo-LA",
    "Lets": "lv-LV",
    "Lithouws": "lt-LT",
    "Macedonisch": "mk-MK",
    "Malagasi": "mg-MG",
    "Maleis": "ms-MY",
    "Maltees": "mt-MT",
    "Maori": "mi-NZ",
    "Marathi": "mr-IN",
    "Mongools": "mn-MN",
    "Nederlands": "nl-NL",
    "Nepali": "ne-NP",
    "Noors": "no-NO",
    "Oekraïens": "uk-UA",
    "Oezbeeks": "uz-UZ",
    "Pools": "pl-PL",
    "Portugees": "pt-PT",
    "Portugees (Brazilië)": "pt-BR",
    "Braziliaans": "pt-BR",
    "Roemeens": "ro-RO",
    "Russisch": "ru-RU",
    "Servisch": "sr-RS",
    "Sloveens": "sl-SI",
    "Spaans": "es-ES",
    "Swahili": "sw-KE",
    "Zweeds": "sv-SE",
    "Thai": "th-TH",
    "Tsjechisch": "cs-CZ",
    "Turks": "tr-TR",
    "Vietnamees": "vi-VN",
    "Wit-Russisch": "be-BY",
    "Zulu": "zu-ZA"
}

LANGUAGES_EN = language_tags_full = {
    "Albanian": "sq-AL",
    "Amharic": "am-ET",
    "Arabic": "ar-SA",
    "Armenian": "hy-AM",
    "Azerbaijani": "az-AZ",
    "Basque": "eu-ES",
    "Bengali": "bn-BD",
    "Bhutanese": "dz-BT",
    "Burmese": "my-MM",
    "Bosnian": "bs-BA",
    "Breton": "br-FR",
    "Bulgarian": "bg-BG",
    "Catalan": "ca-ES",
    "Chinese": "zh-CN",  # Simplified Chinese, for Traditional Chinese it would be "zh-TW"
    "Corsican": "co-FR",
    "Danish": "da-DK",
    "German": "de-DE",
    "Dzongkha": "dz-BT",
    "English": "en-GB",
    "British English": "en-GB",
    "American English": "en-US",
    "American": "en-US",
    "English (US)": "en-US",
    "Estonian": "et-EE",
    "Persian": "fa-IR",
    "Fijian": "fj-FJ",
    "Finnish": "fi-FI",
    "French": "fr-FR",
    "Scottish Gaelic": "gd-GB",
    "Galician": "gl-ES",
    "Georgian": "ka-GE",
    "Greek": "el-GR",
    "Hebrew": "he-IL",
    "Hindi": "hi-IN",
    "Hungarian": "hu-HU",
    "Icelandic": "is-IS",
    "Indonesian": "id-ID",
    "Italian": "it-IT",
    "Japanese": "ja-JP",
    "Javanese": "jv-ID",
    "Kazakh": "kk-KZ",
    "Khmer": "km-KH",
    "Korean": "ko-KR",
    "Croatian": "hr-HR",
    "Lao": "lo-LA",
    "Latvian": "lv-LV",
    "Lithuanian": "lt-LT",
    "Macedonian": "mk-MK",
    "Malagasy": "mg-MG",
    "Malay": "ms-MY",
    "Maltese": "mt-MT",
    "Maori": "mi-NZ",
    "Marathi": "mr-IN",
    "Mongolian": "mn-MN",
    "Dutch": "nl-NL",
    "Nepali": "ne-NP",
    "Norwegian": "no-NO",
    "Ukrainian": "uk-UA",
    "Uzbek": "uz-UZ",
    "Polish": "pl-PL",
    "Portuguese": "pt-PT",
    "Brazilian Portuguese": "pt-BR",
    "Brazilian": "pt-BR",
    "Portuguese (Brazil)": "pt-BR",
    "Romanian": "ro-RO",
    "Russian": "ru-RU",
    "Serbian": "sr-RS",
    "Slovenian": "sl-SI",
    "Spanish": "es-ES",
    "Swahili": "sw-KE",
    "Swedish": "sv-SE",
    "Thai": "th-TH",
    "Czech": "cs-CZ",
    "Turkish": "tr-TR",
    "Vietnamese": "vi-VN",
    "Belarusian": "be-BY",
    "Zulu": "zu-ZA"
}

LANGUAGES = LANGUAGES_NL | LANGUAGES_EN



import json


class Message:
    """ Handles the completion as returned by GPT """
    def __init__(self, role=None, content=None):
        self.role = role
        if not content:
            self.content = None
            self.data = None
        elif isinstance(content, str):
            self.content = content
            self.data = None
        else:
            self.content = json.dumps(content)
            self.data = content

    @classmethod
    def from_dict(cls, data: dict):
        message = cls()
        for key, value in data.items():
            setattr(message, key, value)
        return message

    def __bool__(self):
        return bool(self.content)

    def __str__(self):
        res = f'role: {self.role}'
        res += f' content: {self.content}'
        return res

    def to_dict(self):
        dictionary = {}
        for key, value in self.__dict__.items():
            if value is not None:
                dictionary[key] = value
        return dictionary



import os
import hashlib
import pickle
import re
from pathlib import Path
from lxml import etree
from enum import Enum

from justai.agent.agent import Agent
from justai.tools.log import Log
from justai.tools.prompts import get_prompt, set_prompt_file
from justai.translator.languages import LANGUAGES


# Translation options
class Opt(Enum):
    REPLACE_VARIABLES = 'replace_variables'  # Vervang %variabelen% voor vertaling en zet ze daarna terug
    LOG = 'log'
    STRING_CACHED = 'string_cached'
    CONCATENATED = 'concatenated'  # Of strings zijn samengevoegd met || om ze samen te vertalen
    PROMPT = 'prompt'  # Om een eigen prompt mee te geven.


class Translator(Agent):

    def __init__(self, model=None, **kwargs):
        if not model:
            model = os.environ.get('MODEL', 'claude-3-opus-20240229')
        super().__init__(model, temperature=0, max_tokens=4096, **kwargs)
        set_prompt_file(Path(__file__).parent / 'prompts.toml')
        self.system_message = get_prompt('SYSTEM')
        self.xml = ''
        self.version = ''
        self.logger = Log()

    def load(self, input_file: str | Path):
        with open(input_file, 'r') as f:
            self.read(f.read())

    def read(self, input_string: str):
        # Input bestaat uit <transunit> elementen. Die hebben een datatype property.
        # Binnen elke <transunit> zit een <source> element en komt (na vertaling) een <target> element.
        # ALs datatype == "plaintext" dan zit de te vertalen tekst direct in de <source>
        # Als datatype == "x-DocumentState" dan zit er in de <source> een <g> element met daarin de te vertalen tekst.

        # In 2.0:
        # Input bestaat uit <unit> elementen. Die hebben een Id.
        # Binnen elke <unit> zit een <segment> en daarin een <source>
        # In de source zit ofwel direct tekst, ofwel een <pc> element
        # met daarin nog een <pc> element met daarin de te vertalen tekst
        self.xml = input_string
        self.messages = []
        try:
            self.version = self.xml.split('xliff:document:')[1].split('"')[0].split("'")[0]
        except IndexError:
            raise ValueError('No XLIFF version found in input')
        if self.version not in ['1.2', '2.0']:
            raise ValueError(f'Unsupported XLIFF version: {self.version}')

    def translate(self, language: str, string_cached: bool = False) -> str:
        log = self.logger
        parser = etree.XMLParser(ns_clean=True)
        root = etree.fromstring(self.xml.encode('utf-8'), parser=parser)
        namespaces = {'ns': f'urn:oasis:names:tc:xliff:document:{self.version}'}

        segment_name = 'trans-unit' if self.version == '1.2' else 'segment'

        if self.version >= '2.0':
            root.attrib['trgLang'] = LANGUAGES.get(language)

        # Verzamel teksten als lijst van samengevoegde strings per <source>
        all_texts = []
        translatable = []
        for source in root.xpath('.//ns:source', namespaces=namespaces):
            texts_from_source_element = collect_texts_from_element(source)
            all_texts.append(texts_from_source_element)
            # Verzamel alleen de teksten die ook echt vertaald moeten worden
            translatable_from_source_element = [text for text in texts_from_source_element if is_translatable(text)]
            if translatable_from_source_element:
                translatable.append("||".join(translatable_from_source_element))
        log.info('translate - all_texts', all_texts)
        log.info('translate - translatable_texts', translatable)

        # Vertaal de lijst van met || samengevoegde strings
        options = {Opt.STRING_CACHED: string_cached, Opt.LOG: True, Opt.CONCATENATED: True, Opt.REPLACE_VARIABLES: True}
        translated = self.do_translate(translatable, language, options)

        # Zet nu de vertaalde delen terug in all_texts
        index_in_translated = 0
        for texts_from_source_element in all_texts:
            translatable_from_source_element = [text for text in texts_from_source_element if is_translatable(text)]
            if translatable_from_source_element:
                translated_version = translated[index_in_translated].split("||")

                # Check
                sc = len(translatable_from_source_element)
                tc = len(translated_version)
                if sc != tc:
                    log.error('translate - mismatch',
                              f'Source texts ({sc}) does not match number of translated texts ({tc})')
                    log.info('translate - translatable_from_source_element:', translatable_from_source_element)
                    log.info('translate - translated_version:', translated_version)
                assert sc == tc, 'Mismatch see log'

                for index, text in enumerate(texts_from_source_element):
                    if is_translatable(text):
                        texts_from_source_element[index] = translated_version.pop(0)
                index_in_translated += 1
        flattened_all_texts = [item for sublist in all_texts for item in sublist]  # And flatten

        # Plaats vertaalde teksten in nieuwe <target> elementen met behoud van structuur
        counter = [0]
        for segment in root.xpath(f'.//ns:{segment_name}', namespaces=namespaces):
            source = segment.xpath('.//ns:source', namespaces=namespaces)[0]
            target = etree.SubElement(segment, f'{{urn:oasis:names:tc:xliff:document:{self.version}}}target')
            copy_structure_with_texts(source, target, flattened_all_texts, counter)

        updated_xml = etree.tostring(root, pretty_print=True, xml_declaration=True, encoding='UTF-8').decode('utf-8')
        return updated_xml

    def do_translate(self, texts, language: str, options: dict = {}):
        # Options can be:
        # REPLACE_VARIABLES: True | False
        # LOG: True | False
        # STRING_CACHED: True | False
        # CONCATENATED: True | False
        # PROMPT = 'prompt'  # Om een eigen prompt mee te geven.

        # assert all(is_translatable(text) for text in texts), "Not all translatable"  # !! Tijdelijk
        cache = StringCache(language) if options.get(Opt.STRING_CACHED) else {}
        source_list = [text for text in texts if text not in cache]

        if source_list:
            variables = []
            source_str_no_vars = ''
            for index, text in enumerate(source_list):
                if options.get(Opt.REPLACE_VARIABLES):
                    text_no_vars, v = replace_variables_with_hash(text)
                    variables.extend(v)
                else:
                    text_no_vars = text
                source_str_no_vars += f'{index + 1} [[{text_no_vars}]]\n'
            given_prompt = options.get(Opt.PROMPT)
            if given_prompt:
                prompt = given_prompt.format(language=language, translate_str=source_str_no_vars,
                                             count=len(source_list))
            else:
                prompt = get_prompt('TRANSLATE_MULTIPLE', language=language, translate_str=source_str_no_vars,
                                    count=len(source_list))
            target_str_no_vars = self.chat(prompt, return_json=False, cached=False)
            if options.get(Opt.REPLACE_VARIABLES):
                target_str = replace_hash_with_variables(target_str_no_vars, variables)
            else:
                target_str = target_str_no_vars
            if options.get(Opt.LOG):
                log = self.logger
                log.response('target_str_no_vars', target_str_no_vars)
                log.prompt('prompt', prompt)
                log.info('do_translate - target_str', target_str)

            target_list = [t.split(']]')[0] for t in target_str.split('[[')[1:]]

            count = 1
            for source, translation in zip(source_list, target_list):

                if options.get(Opt.CONCATENATED):
                    source_parts = source.split('||')
                    translation_parts = translation.split('||')
                    # Check op het aantal onderdelen tussen ||. Dit moet gelijk zijn in de bron en de vertaling
                    sc = len(source_parts)
                    tc = len(translation_parts)
                    if tc != sc:
                        log.warning('do_translate',
                                    f'Number of translated texts ({tc}) does not match number of source texts ({sc}). Correcting.')
                        # Model krijgt het niet voor elkaar om een zin met evenveel delen terug te geven
                        # We vertalen de stukjes los van elkaar.
                        translation_parts = self.do_translate(source_parts, language, options)
                        # if tc < sc:
                        #     translation_parts += [' '] * (sc - tc)
                        # else:
                        #     # Dit is een hack! Het model geeft kennelijk meer ||'s terug in dan in het origineel.
                        #     # Dat breekt de code verderop. Daarom voegen we de laatste onderdelen samen.
                        #     # Maar dat levert wel een onjuiste vertaling op.
                        #     translation_parts = translation_parts[:sc - 1] + [' '.join(translation_parts[sc - 1:])]
                        sc = len(source_parts)
                        tc = len(translation_parts)
                        if tc != sc:
                            log.error('do_translate',
                                      f'Number of translated texts ({tc}) still does not match number of source texts ({sc}). Correcting.')
                        assert (tc == sc), 'Mismatch in number of parts, see log for info'
                else:
                    source_parts = [source]
                    translation_parts = [translation]

                # Zorg dat de vertaling dezelfde whitespace aan het begin en eind heeft als de bron
                for i, (source_part, translation_part) in enumerate(zip(source_parts, translation_parts)):
                    if source_part.strip() and (source_part[0] == ' ' or source_part[-1] == ' '):
                        start_spaces = (len(source_part) - len(source_part.lstrip(' '))) * ' '
                        end_spaces = (len(source_part) - len(source_part.rstrip(' '))) * ' '
                        translation_parts[i] = start_spaces + translation_part.strip() + end_spaces

                target_list[count - 1] = '||'.join(translation_parts)

                count += 1

                if options.get(Opt.LOG):
                    ratio = len(source) / len(translation)
                    if ratio >= 1.5 or ratio <= 0.7:
                        log.warning('', f'Vertaling van {source} naar {translation} is onverwacht lang of kort')

            translation_dict = dict(zip(source_list, target_list))
            cache.update(translation_dict)
            if options.get(Opt.STRING_CACHED):
                cache.save()

        translations = [cache.get(text, text) for text in texts]

        if options.get(Opt.CONCATENATED):
            for s, t in zip(translations, texts):
                sparts, tparts = s.split('||'), t.split('||')
                if len(sparts) != len(tparts):
                    assert False, f'Number of parts in source ({len(sparts)}) does not match number of parts in translation ({len(tparts)})'

        return translations

    def translate_dict(self, text_dict, language: str, options):
        return {k: v for k, v in zip(text_dict.keys(),
                                     self.do_translate(list(text_dict.values()), language, options))}


def replace_variables_with_hash(text):
    # Vindt alle variabelen in de tekst
    variables = re.findall(r'%[^%]+%', text)
    # Vervang alle variabelen in de tekst met ###
    # Het model heeft moeite met newlines. Daarom vervangen we ze door @@ en na vertaling weer terug.
    modified_text = re.sub(r'%[^%]+%', '###', text).replace('\n', '@@')
    return modified_text, variables


def replace_hash_with_variables(text, variables):
    for variable in variables:
        text = text.replace('###', variable, 1)
    # en zet de newlines terug
    text = text.replace('@@', '\n')
    return text


def collect_texts_from_element(element):
    texts = []
    # if element.text and element.text.strip():
    #    texts.append(element.text.strip())
    if element.text:
        texts.append(element.text)
    for child in element:
        texts.extend(collect_texts_from_element(child))
    return texts


def copy_structure_with_texts(source, target, translated_texts, counter=[0], log=None):
    """ Kopieer de structuur van <source> naar <target> en behoud de teksten """
    if source.text:  # and source.text.strip():
        try:
            target.text = translated_texts[counter[0]]
            counter[0] += 1
        except IndexError:
            if log:
                log.error('IndexError', 'IndexError in copy_structure_with_texts')
    for child in source:
        child_copy = etree.SubElement(target, child.tag, attrib=child.attrib)
        copy_structure_with_texts(child, child_copy, translated_texts, counter, log)


def is_translatable(text) -> bool:
    """ Returns True if the unit should be translated """
    return text and re.search('[a-zA-Z]{2}', text) and text[0] not in ('%', '<')


def split_list_in_sublists(source_list, max_chunk_len):
    chunks = []
    for text in source_list:
        if not chunks or chunks[-1] and len(chunks[-1]) + len(text) > max_chunk_len:
            chunks.append([text])
        else:
            chunks[-1].append(text)
    return chunks


class StringCache:
    def __init__(self, language: str):
        self.language = language
        self.cache = {}
        self.file = Path(__file__).parent / (self.language + '.pickle')
        try:
            with open(self.file, 'rb') as f:
                self.cache = pickle.load(f)
        except (FileNotFoundError, EOFError):
            self.cache = {}

    def get(self, source, default=None):
        key = self.get_key(source)
        return self.cache.get(key, default)

    def set(self, source, translation):
        key = self.get_key(source)
        self.cache[key] = translation

    def __contains__(self, source):
        key = self.get_key(source)
        return key in self.cache

    def __len__(self):
        return len(self.cache)

    def update(self, translation_dict):
        for source, translation in translation_dict.items():
            self.set(source, translation)

    def save(self):
        with open(self.file, 'wb') as f:
            pickle.dump(self.cache, f)

    def clear(self):
        self.cache = {}
        self.save()

    @classmethod
    def get_key(cls, source):
        return hashlib.md5(source.encode('utf-8')).hexdigest()


def parse_xliff_with_unit_clusters(xliff_content, max_chunk_size):
    # Functie alleen gebruikt voor tests. Deze is gelijk aan de parseXLIFFWithUnitClusters uit javascript
    # en is bedoeld om de real life situatie met het splitsen van de XLIFF in clusters te simuleren.

    # Bepaal de versie van xliffContent
    version_match = re.search(r'<xliff[^>]*\s+version="([0-9.]+)"', xliff_content)
    version = version_match.group(1) if version_match else None

    # Splits xliff in header, clusters van maxChunkSize, and footer
    header_re = r'^(.*?)<unit ' if version == '2.0' else r'^(.*?)<trans-unit '
    header_match = re.match(header_re, xliff_content, re.DOTALL)
    header = header_match.group(1) if header_match else None

    # Extract units and cluster them
    units = []
    cluster = ''
    unit_re = r'<unit .*?</unit>' if version == '2.0' else r'<trans-unit .*?</trans-unit>'
    matches = re.findall(unit_re, xliff_content, re.DOTALL)

    for match in matches:
        unit = match
        if len(unit) > max_chunk_size:
            # If current unit is larger than maxChunkSize, push current cluster (if not empty) and then this unit
            if cluster:
                units.append(cluster)
                cluster = ''
            units.append(unit)
        elif len(cluster) + len(unit) > max_chunk_size:
            # If adding this unit exceeds the limit, push current cluster and start a new one
            units.append(cluster)
            cluster = unit
        else:
            # Add this unit to current cluster
            cluster += unit

    # Don't forget to add the last cluster if it exists
    if cluster:
        units.append(cluster)

    # Extract footer
    footer_re = r'</unit>' if version == '2.0' else r'</trans-unit>'
    footer = xliff_content.split(footer_re)[-1]

    return {"header": header, "units": units, "footer": footer, "version": version}



""" Handles the GPT API and the conversation state. """
import json
import os
import re
import time
from pathlib import Path

from justai.tools.cache import cached_llm_response
from justai.tools.display import print_message, color_print, SYSTEM_COLOR
from justai.agent.message import Message
from justai.models.modelfactory import ModelFactory


class Agent:
    def __init__(self, model_name: str, **kwargs):
        
        # Model parameters
        self.model = ModelFactory.create(model_name, **kwargs)

        # Parameters to save the current conversation
        self.save_dir = Path(__file__).resolve().parent / 'saves'
        self.message_memory = 20  # Number of messages to remember. Limits token usage.
        self.messages = []  # List of Message objects

        self.input_token_count = 0
        self.output_token_count = 0
        self.last_response_time = 0
        
        self.logger = None
        
    def __setattr__(self, name, value):
        if name not in self.__dict__ and hasattr(self, 'model') and name in self.model.model_params:
            # Not an existing property model but a model_params property. Set it in model_params
            self.model.model_params[name] = value
        else:
            # Update the property as intended
            super().__setattr__(name, value)

    @classmethod
    def from_json(cls, model_name, model_data, **kwargs):
        """ Creates an agent from a json string. Usefull in stateless environments like a web page """
        agent = cls(model_name, **kwargs)
        dictionary = json.loads(model_data)
        for key, value in dictionary.items():
            match key:
                case 'save_dir':
                    agent.save_dir = Path(value)
                case 'messages':
                    agent.messages = [Message.from_dict(m) for m in value]
                case _:
                    agent.__setattr__(key, value)
        return agent

    def set_api_key(self, key: str):
        """ Used when using Aigent from a browser where the user has to specify a key """
        self.model.set('api_key', key)

    @property
    def system(self):  # This function can be overwritten by child classes to make the system message dynamic
        return self.model.system_message

    @system.setter
    def system(self, value):
        self.model.system_message = value

    def reset(self):
        self.messages = []

    def get_messages(self) -> list[dict]:
        result = [{'role': 'system', 'content': self.system}]
        for m in self.messages[-self.message_memory:]:
            message = {'role': m.role, 'content': m.content}
            result.append(message)
        return result

    def last_token_count(self):
        return self.input_token_count, self.output_token_count, self.input_token_count + self.output_token_count

    def chat(self, prompt, return_json=False, cached=True):
        start_time = time.time()
        self.messages.append(Message('user', prompt))
        model_response = cached_llm_response(self.model, self.get_messages(), return_json=return_json, use_cache=cached)
        text, self.input_token_count, self.output_token_count = model_response
        self.messages.append(Message('assistant', text))
        self.last_response_time = time.time() - start_time
        return text
    
    async def chat_async(self, prompt):
        self.messages.append(Message('user', prompt))
        for answer_text in self.model.chat_async(messages=self.get_messages()): 
            if answer_text:
                yield answer_text

    def after_response(self):
        # content is in messages[-1]['completion']['choices'][0]['message']['content']
        return  # Can be overridden

    def save(self, name=None):
        if not name:
            name = re.sub(r'\W+', '', self.messages[0].text).replace(' ', '_')[:20]
        self.save_dir.mkdir(exist_ok=True)
        with open((self.save_dir / name).with_suffix('.txt'), "w") as f:
            f.write(f"system: {self.system}\n")
            for message in self.messages:
                f.write(f"{message.role}: {message.text}\n")

    def load(self, name: str):
        def save_message(msg):
            if msg.role == 'system':
                self.system = msg.content
            else:
                self.messages += [msg]

        self.messages = []
        if not name.endswith('.txt'):
            name += '.txt'
        filename = self.save_dir / name
        if not os.path.isfile(filename):
            color_print(f"New conversation:  {filename}", color=SYSTEM_COLOR)
            return
        with open(filename, "r") as f:
            message = Message()
            assert not message
            for line in f.readlines():
                line = line[:-1]
                try:
                    role, text = line.split(': ', 1)
                except ValueError:
                    message.content += '\n' + line
                    continue
                if role in ['system', 'user', 'assistant', 'function']:
                    if message:
                        save_message(message)
                    message = Message(role=role, content=text)
                else:
                    message.content += '\n' + line
            if message:
                save_message(message)
        print_message(Message('system', self.system()), 'system')
        for message in self.messages:
            print_message(message.text, message.role)

    def file_input(self, filename):
        with open(filename, "r") as f:
            prompt = f.read()
        self.chat(prompt)

    def dumps(self) -> str:
        data = {}
        for key, value in self.__dict__.items():
            if value is not None:
                try:
                    json.dumps(value)
                    data[key] = value
                except (TypeError, ValueError):
                    match key:
                        case 'save_dir':
                            data[key] = str(value)
                        case 'messages':
                            data[key] = [message.to_dict() for message in value]
        return json.dumps(data, indent=2)

    def token_count(self, text: str):
        return self.model.token_count(text)



from justai.models.model import Model


class ModelFactory:
    @staticmethod
    def create(model_name: str, **kwargs) -> Model:
        if model_name.startswith("gpt"):
            from justai.models.openai_models import OpenAIModel
            return OpenAIModel(model_name, params=kwargs)
        elif model_name.endswith(".gguf"):
            from justai.models.gguf_models import GuffModel
            return GuffModel(model_name, params=kwargs)
        elif model_name.startswith("claude"):
            from justai.models.anthropic_models import AnthropicModel
            return AnthropicModel(model_name, params=kwargs)
        else:
            raise ValueError(f"Model {model_name} not supported")



import json
import os

import openai
import tiktoken
from dotenv import load_dotenv
from openai import APIConnectionError, APIError, RateLimitError, OpenAI
from openai._types import NOT_GIVEN

from justai.tools.display import color_print, ERROR_COLOR, DEBUG_COLOR1, DEBUG_COLOR2, SYSTEM_COLOR
from justai.models.model import Model


class OpenAIModel(Model):
    def __init__(self, model_name: str, params: dict = None):
        system_message = f"You are {model_name}, a large language model trained by OpenAI."
        super().__init__(model_name, params, system_message)

        # Authentication
        if "OPENAI_API_KEY" in params:
            openai.api_key = params["OPENAI_API_KEY"]
        else:
            if not os.getenv("OPENAI_API_KEY"):
                load_dotenv()  # Load the .env file into the environment
            openai.api_key = os.getenv("OPENAI_API_KEY")
        if not openai.api_key:
            color_print("No OpenAI API key found. Create one at https://platform.openai.com/account/api-keys and " +
                        "set it in the .env file like OPENAI_API_KEY=here_comes_your_key.", color=ERROR_COLOR)
        self.client = OpenAI()

        # The maximum number of tokens to generate in the completion.
        # Defaults to 16
        # The token count of your prompt plus max_tokens cannot exceed the model's context length.
        # Most models have a context length of 2048 tokens (except for the newest models, which support 4096).
        self.model_params['max_tokens'] = params.get('max_tokens', 800)

        # What sampling temperature to use, between 0 and 2.
        # Higher values like 0.8 will make the output more random, while lower values like 0.2
        # will make it more focused and deterministic.
        # We generally recommend altering this or top_p but not both
        # Defaults to 1
        self.model_params['temperature'] = params.get('temperature', 0.5)

        # An alternative to sampling with temperature, called nucleus sampling,
        # where the model considers the results of the tokens with top_p probability mass.
        # So 0.1 means only the tokens comprising the top 10% probability mass are considered.
        # We generally recommend altering this or temperature but not both.
        # Defaults to 1
        self.model_params['top_p'] = params.get('top_p', 1)

        # How many completions to generate for each prompt.
        # Because this parameter generates many completions, it can quickly consume your token quota.
        # Use carefully and ensure that you have reasonable settings for max_tokens and stop.
        self.model_params['n'] = params.get('n', 1)

        # Up to 4 sequences where the API will stop generating further tokens.
        # The returned text will not contain the stop sequence.
        # Example: [" Human:", " AI:"]
        self.model_params['stop'] = params.get('stop')

        # Number between -2.0 and 2.0.
        # Positive values penalize new tokens based on whether they appear in the text so far,
        # increasing the model's likelihood to talk about new topics.
        # Defaults to 0
        self.model_params['presence_penalty'] = params.get('presence_penalty', 0)

        # Number between -2.0 and 2.0.
        # Positive values penalize new tokens based on their existing frequency in the text so far,
        # decreasing the model's likelihood to repeat the same line verbatim.
        # Defaults to 0
        self.model_params['frequency_penalty'] = params.get('frequency_penalty', 0)

    def chat(self, messages: list[dict], return_json: bool, use_cache: bool = False, max_retries=None):
        if max_retries is None:
            max_retries = 3

        if self.debug:
            color_print("\nRunning completion with these messages", color=DEBUG_COLOR1)
            [color_print(m, color=DEBUG_COLOR1) for m in messages if hasattr(m, 'text')]
            print()

        last_error = None
        for _ in range(max_retries):
            try:
                completion = self.completion(messages, return_json)
                message_text = completion.choices[0].message.content
                input_token_count = completion.usage.prompt_tokens
                output_token_count = completion.usage.completion_tokens
                break
            except APIConnectionError as e:
                color_print("Connection error.", color=SYSTEM_COLOR)
                last_error = e
            except APIError as e:
                color_print("API error", color=SYSTEM_COLOR)
                last_error = e
            except RateLimitError as e:
                color_print(f"{self.model_name} is overloaded", color=SYSTEM_COLOR)
                last_error = e
        else:
            print('Too many errors. Aborting.')
            raise last_error

        if message_text.startswith('```json'):
            message_text = message_text[7:-3]
        if self.debug:
            color_print(f"{message_text}", color=DEBUG_COLOR2)

        text = json.loads(message_text) if return_json else message_text
        return text, input_token_count, output_token_count
    
    def chat_async(self, messages: list[dict]):
        for item in self.completion(messages, stream=True):
            if hasattr(item.choices[0].delta, "content"):
                yield item.choices[0].delta.content
               
    def completion(self, messages: list[dict], return_json: bool = False, stream: bool=False):
        return self.client.chat.completions.create(
            model=self.model_name,
            messages=messages,
            temperature=self.model_params['temperature'],
            max_tokens=self.model_params['max_tokens'],
            n=self.model_params['n'],
            top_p=self.model_params['top_p'],
            frequency_penalty=self.model_params['frequency_penalty'],
            presence_penalty=self.model_params['presence_penalty'],
            stop=self.model_params['stop'],
            response_format={"type": "json_object"} if return_json else NOT_GIVEN,
            stream = stream
        )

    def token_count(self, text: str) -> int:
        encoding = tiktoken.encoding_for_model(self.model_name)
        return len(encoding.encode(text))



from justai.agent.agent import Agent
from justai.translator.translator import Translator
from justai.interactive.repl import Repl
from justai.interactive.commands import CommandHandler
from justai.tools.prompts import get_prompt, set_prompt_file, add_prompt_file
from justai.tools.log import Log, set_log_dir

if __name__ == '__main__':
    # Ondertaande om de voorkomen dat import optimizer ze leeg gooit
    a = Agent
    t = Translator
    r = Repl
    c = CommandHandler
    g = get_prompt
    s = set_prompt_file
    a = add_prompt_file
    l = Log
    s = set_log_dir
    
